#ident "$Id: dpy_x11.c,v 1.2 2000/12/12 10:34:24 pierre Exp $"
/*
 * OMM Display implementation for X11 library
 */
#include "omm.h"
#include <fcntl.h>
#include <X11/Xlib.h>

#include "dpy.h"

static pthread_mutex_t write_mutex;
static GC gc;
static Display *display;
static int screen;
static Window  win, root;
static unsigned long white_pixel, black_pixel;
static char mapped, exposed;
static int font_height, font_width, font_descent;
static boolean x11_use_stdin;
static Window old_focus_win;
static int old_revert_to;
static XColor green1, green2, blue;

/* Init */
void lcd_hw_init (boolean use_stdin)
{
  XFontStruct *xfs;
  //  XSetWindowAttributes wa;
  XColor color;
  char *owin;

  pthread_mutex_init (&write_mutex, NULL);

  if (!XInitThreads ()) {
    log_err ("Threads not supported, exiting.");
    exit (1);
  }

  if ((display = XOpenDisplay (NULL)) == NULL) {
    log_err ("Can't open display, exiting.");
    exit (1);
  }

  screen = DefaultScreen (display);

  green1.red = 0x78 << 8;
  green1.green = 0xb8 << 8;
  green1.blue = 0;
  XAllocColor (display, DefaultColormap (display, screen), &green1);

  green2.red = 0x90 << 8;
  green2.green = 0xd8 << 8;
  green2.blue = 0;
  XAllocColor (display, DefaultColormap (display, screen), &green2);

  gc = DefaultGC (display, screen);
  
  /* Set font */
  xfs = XLoadQueryFont (display, "10x20");
  if (!xfs) {
    log_err ("Can't load font, exiting.");
    exit (1);
  }
  font_width = xfs->max_bounds.width;
  font_height = xfs->ascent + xfs->descent;
  font_descent = xfs->descent;
  XSetFont (display, gc, xfs->fid);
  log_debug (DEBUG_GENERAL, "Font %d x %d", font_width, font_height);

  root = RootWindow (display, screen);
  white_pixel = WhitePixel (display, screen);
  black_pixel = BlackPixel (display, screen);

  XSynchronize (display, True);

  /* Use existing win */
  if ((owin = getenv("OMMWIN"))) {
    unsigned long x;

    sscanf (owin, "%x", &x);
    win = (Window)x;
  }
  else {
    /*
    wa.background_pixel = green2.pixel;
    wa.border_pixel = black_pixel;
    wa.override_redirect = True;
    wa.event_mask = ExposureMask|KeyPressMask;
    win = XCreateWindow (display, root, 40, 50, LCD_WIDTH * (font_width+2), LCD_HEIGHT * (font_height+2), 4, CopyFromParent, CopyFromParent, CopyFromParent, CWBackPixel|CWBorderPixel|CWOverrideRedirect|CWEventMask, &wa);
    */
    win = XCreateSimpleWindow (display, root, 0, 0, LCD_WIDTH * (font_width+2), LCD_HEIGHT * (font_height+2), 1, black_pixel, green2.pixel);
    XSelectInput (display, win, ExposureMask|KeyPressMask);
    XMapWindow (display, win);
  }

  x11_use_stdin = use_stdin;
  if (use_stdin) {
    XGetInputFocus (display, &old_focus_win, &old_revert_to);
    XSetInputFocus (display, win, RevertToNone, CurrentTime);
  }
}

/* Write string */
void lcd_hw_write (int x, int y, char *s)
{
  XWindowAttributes xwa;
  register int i;

  pthread_mutex_lock (&write_mutex);

  if (!mapped && !exposed) {
    for (i = 0 ; i < 10 ; i++) {
      XGetWindowAttributes (display, win, &xwa);
      if (xwa.map_state == IsUnmapped) {
	log_debug (DEBUG_GENERAL, "Window is not mapped (%d)!", i);
	usleep (100000);
	continue;
      }
      else { 
	mapped = 1;
	break;
      }
    }
  }

  XSetBackground (display, gc, green1.pixel);
  for (i = 0 ; i < strlen(s) ; i++) {
    int xx, yy;

    xx = 1+ (x+i) * (font_width+2);
    yy = 1+ (y+1) * (font_height+2) - font_descent;

    //    log_debug (DEBUG_GENERAL, "lcd_hw_write: %d/%d %c", xx, yy, *(s+i)); 
    XDrawImageString (display, win, gc, xx,  yy, s+i, 1) ;
  }

  //  usleep (1000);

  pthread_mutex_unlock (&write_mutex);
}

/* Close */
void lcd_hw_close ()
{
  /* Give back focus */
  if (x11_use_stdin)
    XSetInputFocus (display, old_focus_win, old_revert_to, CurrentTime);
}

/* Get input */
int lcd_hw_input ()
{
  char buf[2], c;
  KeySym ks;
  XEvent ev;
  fd_set rmask;
  int ntoread, nfds;
  int fd_x11 = ConnectionNumber(display);

  FD_ZERO (&rmask);	
  FD_SET (fd_fifo, &rmask);
  if (x11_use_stdin)
    FD_SET (fd_x11, &rmask);

  XFlush (display);
  nfds = (fd_x11 > fd_fifo) ? fd_x11 + 1 : fd_fifo + 1;
  ntoread = select (nfds, &rmask, NULL, NULL, NULL);

  if (ntoread < 0) {
    log_err ("lcd_hw_input: select < 0 (%s) !\n", strerror(errno));
    return -1;
  }
  else if (ntoread == 0) {
    log_err ("lcd_hw_input: select = 0 !\n");
    return -1;
  }
  else { /* ntoread > 0 */
    if (FD_ISSET (fd_x11, &rmask)) {
      while (XCheckMaskEvent (display, ~0, &ev)) {
	switch (ev.type) {
	case Expose :
	  log_debug (DEBUG_GENERAL, "Expose !");
	  exposed = 1;
	  break;
	case KeyPress :
	  if (XLookupString ((XKeyEvent *)&ev, &buf[0], 1, &ks, 0)) {
	    return buf[0];
	  }
	  else {
	    log_err ("lcd_hw_input: XLookupString error !");
	    return -1;
	  }

	default:
	  return -1;
	}
      }
    }
   
    if (FD_ISSET (fd_fifo, &rmask)) {
      if (read (fd_fifo, &c, 1) < 0) {
	log_err ("lcd_hw_input: read error on %d", fd_fifo);
	return -1;
      }

      return c;
    }
  }
}

/* Get win */
unsigned long lcd_hw_get_window ()
{
  return (unsigned long)win;
}
