#ident "$Id: lcd.c,v 1.2 2000/12/12 10:36:21 pierre Exp pierre $"
/*
 * OMM Display functions. 
 *
 * This level is independent from hardware.
 */
#include "omm.h"

static chardef screen_def[LCD_HEIGHT][LCD_WIDTH];
static chardef screen_save[LCD_HEIGHT][LCD_WIDTH];
static pthread_t blink_thread;
static int hscroll_index;
static int hscroll_direction;	/*eb 1 = L->R ; -1 = R->L eb*/
static int hscroll_init;		/*eb 1 = scrolling init eb*/
static int x_lcd, y_lcd;
static char screen_state = LCD_FG_STATE;

/* Handle blinking strings */
static void *blink_process (void *dummy) 
{
  register int x, y, i;
  int count = 0;
  char str[LCD_WIDTH+1];

  while (1) {
    for (y = 0 ; y < LCD_HEIGHT ; y++) {
      for (x = 0 ; x < LCD_WIDTH ; x++) {
	if (screen_def[y][x].attr & ATTR_BLINK) {
	  /* Build a blinking string and display it */
	  i = 0;
	  while (i < LCD_WIDTH && (screen_def[y][x+i].attr & ATTR_BLINK)) {
	    str[i] = (count % 2 ? ' ' : screen_def[y][x+i].code);
	    i++;
	  }

	  str[i] = 0;
	  if (screen_state == LCD_FG_STATE)
	    lcd_hw_write (x, y, str);
	}
      }
    }

    count++;
    usleep (500000);
  }
}  

/* Update the background screen image */
static void update_screen (int x, int y, char *s, lcd_attr attr)
{
  register char *p = s;
  
  for (p = s ; *p ; p++) {
    screen_def[y][x].code = *p;
    screen_def[y][x++].attr = attr;
  }
}

/* For test only: dump the screen image to log file */
void lcd_dump_screen ()
{
  register int x, y;
  char line_def[4*LCD_WIDTH+1];

  for (y = 0 ; y < LCD_HEIGHT ; y++) {
    memset ((void*)line_def, 0, sizeof(line_def));
    for (x = 0 ; x < LCD_WIDTH ; x++) {
      sprintf (&line_def[x << 2], "%c/%d ", screen_def[y][x].code, screen_def[y][x].attr);
    }

    log_debug (DEBUG_GENERAL, "%s", line_def);
  }
}

/* Clear the screen and hide it (background mode) */
void lcd_bg_screen ()
{
  register int y;
  char buf[LCD_WIDTH+1];

  log_debug (DEBUG_GENERAL, "lcd_bg_screen");
  screen_state = LCD_BG_STATE;

  for (y = 0 ; y < LCD_HEIGHT ; y++) {
    memset ((void*)buf, ' ', sizeof(buf)-1);
    buf[LCD_WIDTH] = 0;
    lcd_hw_write (0, y, buf);
  }

  /* Send a reply to say the job is done */
  write (fd_fifo_reply, "0", 1);

  log_debug (DEBUG_GENERAL, "lcd_bg_screen: OK");
}

/* Back to foreground mode */
void lcd_fg_screen ()
{
  register int x, y;
  char buf[LCD_WIDTH+1];

  log_debug (DEBUG_GENERAL, "lcd_fg_screen");
  screen_state = LCD_FG_STATE;

  for (y = 0 ; y < LCD_HEIGHT ; y++) {
    memset ((void*)buf, ' ', sizeof(buf)-1);
    buf[LCD_WIDTH] = 0;
    for (x = 0 ; x < LCD_WIDTH ; x++)
      buf[x] = screen_def[y][x].code;
      
    log_debug (DEBUG_GENERAL, "%s", buf);
    lcd_hw_write (0, y, buf);
  }

  /* Send a reply to say the job is done */
  write (fd_fifo_reply, "0", 1);

  log_debug (DEBUG_GENERAL, "lcd_fg_screen: OK");
}

/* Save/restore parts of the screen */
void lcd_save_area (int x1, int y1, int x2, int y2)
{
  register int x, y;

  for (y = y1 ; y <= y2 ; y++) {
    for (x = x1 ; x <= x2 ; x++) {
      screen_save[y][x].code = screen_def[y][x].code;
      screen_save[y][x].attr = screen_def[y][x].attr;
    }
  }
}

void lcd_restore_area (int x1, int y1, int x2, int y2)
{
  char buf[2];
  register int x, y;

  buf[1] = 0;

  for (y = y1 ; y <= y2 ; y++) {
    for (x = x1 ; x <= x2 ; x++) {
      if (screen_save[y][x].code != screen_def[y][x].code || screen_save[y][x].attr != screen_def[y][x].attr) {
	buf[0] = screen_save[y][x].code;
	lcd_write (x, y, buf, screen_save[y][x].attr);
      }
    }
  }
}


/* Init function */
void lcd_init (boolean use_stdin) 
{
  memset ((void*)screen_def, 0, (size_t)(LCD_HEIGHT*LCD_WIDTH*sizeof(chardef)));
  lcd_hw_init (use_stdin);

  /* Init thread attributes */
  if (pthread_attr_setdetachstate (&thread_attr, PTHREAD_CREATE_DETACHED) != 0)
    log_err ("pthread_attr_setdetachstate: %s", strerror (errno));

  if (pthread_create (&blink_thread, &thread_attr, blink_process, NULL) != 0)
    log_err ("pthread_create error for blink_process");
}

/* Close function */
void lcd_close () 
{
  lcd_hw_close ();
}

/* Get input function */
int lcd_input ()
{
  return lcd_hw_input ();
}

/* 
 * General write string function: 
 * x is column from 0 to LCD_WIDTH-1
 * y is line from 0 to LCD_HEIGHT-1
 */
void lcd_write (int x, int y, char *s, lcd_attr attr)
{
  if (!s)
    return;

  if (x < 0 || x > LCD_WIDTH-1 || y < 0 || y > LCD_HEIGHT-1)
    return;

  x_lcd = x;
  y_lcd = y;

  update_screen (x, y, s, attr);
  if (screen_state == LCD_FG_STATE)
    lcd_hw_write (x, y, s);
}

void lcd_move (int x, int y)
{
  if (x < 0 || x > LCD_WIDTH-1 || y < 0 || y > LCD_HEIGHT-1)
    return;

  x_lcd = x;
  y_lcd = y;
}

/* Next functions are derived from lcd_write() */
void lcd_addch (int c, lcd_attr attr)
{
  char str[2];

  str[0] = c;  
  str[1] = 0;
  lcd_write (x_lcd, y_lcd, str, attr);
  x_lcd++;
}

void lcd_addstr (char *s, lcd_attr attr)
{
  lcd_write (x_lcd, y_lcd, s, attr);
  x_lcd += strlen(s);
  if (x_lcd > LCD_WIDTH-1)
    x_lcd = LCD_WIDTH-1;
}

void lcd_write_center (int y, char *s, lcd_attr attr)
{
  register int x;
  
  x = (LCD_WIDTH - strlen (s))/2;

  /* anything to clear before string ? */
  if (x)
    lcd_clear (0, y, x-1, y);

  /* after ? */
  if (x+strlen(s) < LCD_WIDTH-1)
    lcd_clear (x+strlen(s), y, LCD_WIDTH-1, y);
  
  lcd_write (x, y, s, attr);
}



void lcd_clear (int x1, int y1, int x2, int y2)
{
  register int x, y;

  for (y = y1 ; y <= y2 ; y++) 
    for (x = x1 ; x <= x2 ; x++) {
      if (screen_def[y][x].code != ' ')
	lcd_write (x, y, " ", ATTR_NORMAL);
    }
}

void lcd_clear_line (int y)
{
  char blank_line[LCD_WIDTH+1];

  memset ((void*)blank_line, ' ', LCD_WIDTH);
  blank_line[LCD_WIDTH] = 0;

  lcd_clear (0, y, LCD_WIDTH-1, y);
  //lcd_write (0, y, blank_line, ATTR_NORMAL);
}

void lcd_clear_screen ()
{
  register int y;

  for (y = 0 ; y < LCD_HEIGHT ; y++)
    lcd_clear_line (y);
}


void lcd_init_hscroll ()
{
  hscroll_direction = 1;
  hscroll_init = 1;
  hscroll_index = 0;
}


void lcd_write_hscroll (int x, int y, char *s)
{
  char buf[LCD_WIDTH+1];
  int max_hscroll, max_width;

  if (!s || *s == 0)
    s = "No info!";

  /* centered if x<0 */
  max_width = (x < 0 ? LCD_WIDTH : LCD_WIDTH-x);
  max_hscroll = strlen (s) - max_width;

  /* Horizontal scrolling ? */
  if (max_hscroll > 0) {
    strncpy (buf, s+hscroll_index, max_width);
    buf[max_width] = 0;

    /* Scroll the title and wrap it */
    /*eb added direction support for better scrolling eb*/
    if (hscroll_index < max_hscroll) {
      if (hscroll_direction == 1)
	hscroll_index++;
      else {
	if (hscroll_index > 0)
	  hscroll_index--;
	else
	  hscroll_direction = 1;
      }
    }
    else
      {
     	hscroll_direction = -1;
     	hscroll_index--;
      }
  }
  else {
    strcpy (buf, s);
    hscroll_index++;
  }

  /*eb Clear the line before display only the 1rst time eb*/
  if (hscroll_init-- == 1) {
    if (x < 0)
      lcd_write_center (y, buf, ATTR_NORMAL);
    else
      lcd_write (x, y, buf, ATTR_NORMAL);
    hscroll_index++;
  }

  if (max_hscroll > 0) {
    if (x < 0)
      lcd_write_center (y, buf, ATTR_NORMAL);
    else
      lcd_write (x, y, buf, ATTR_NORMAL);
  }
}

/* Get window */
unsigned long lcd_get_window ()
{
  return lcd_hw_get_window ();
}
