#ident "$Id$"
/*
 * OMM Napster client
 */

#define EXTERN
#include "omm.h"
#include "napster.h"

static char *omm_dir = DEFAULT_OMMDIR;
static int nap_pid;
static int pipe_out[2];
static int fd_napio;
static pthread_t read_nap_thread;
static pthread_t scrolled_thread;
static char *scrolled_str;
static char song_name[1024];
static pthread_mutex_t scroll_mutex;

/* States handling definition */
static omm_event_handler napster_event_handlers[] = {
  'q', OMM_STATE_INIT, napster_quit,		/* Quit */
  's', OMM_STATE_INIT, napster_search,
  'j', OMM_STATE_INIT, napster_next_song,
  'k', OMM_STATE_INIT, napster_prev_song,
  'B', OMM_STATE_INIT, lcd_bg_screen,
  'F', OMM_STATE_INIT, lcd_fg_screen,
  OMM_DEFAULT_EVENT, OMM_ANY_STATE, napster_default_handler
};

static void got_interrupt (int sig)
{	
  log_debug (DEBUG_NAPSTER, "Got interrupt %d!", sig);

  napster_quit ();
  exit (1);
}

static void send_nap (char *s)
{
  log_debug (DEBUG_NAPSTER, "send_nap: %s", s);
  write (NAP_OUT, s, strlen(s));
}


static void *display_scrolled_process (void *dummy)
{
  for (;;) {
    pthread_mutex_lock (&scroll_mutex);
    if (scrolled_str)
      lcd_write_hscroll (-1, 1, scrolled_str);
    pthread_mutex_unlock (&scroll_mutex);
    usleep (500000);

  }
}

/* Next/prev song */
static void napster_next_song ()
{
  send_nap ("j");
}

static void napster_prev_song ()
{
  send_nap ("k");
}

/* Quit Napster */
static void napster_quit ()
{
  if (nap_pid)
    kill (nap_pid, SIGTERM);
  close (fd_napio);

  omm_state = OMM_STATE_INIT;
  omm_api_close ();
}

void input_ok ()
{
  char buf[256], *s;

  s = omm_api_get_input_str();

  log_debug (DEBUG_NAPSTER, "input OK, str= '%s'", s);

  sprintf (buf, "/search %s\n", s);
  send_nap (buf);
}

void input_aborted ()
{
  lcd_restore_area (0, 1, LCD_WIDTH-1, 3);
  log_debug (DEBUG_NAPSTER, "input ABORTED!");
}

/* Search */
static void napster_search ()
{
  lcd_save_area (0, 1, LCD_WIDTH-1, 3);
  omm_api_start_input ("Search ?", input_ok, input_aborted);
}


static void read_nap_fifo (char *s, int size)
{
  char c = 0;
  int cnt = 0;

  memset ((void*)s, 0, size);
  while (c != '\n' && cnt < size) {
    if (read (fd_napio, &c, 1) < 0) {
      log_err ("read_nap_fifo: %s: %s", s, strerror(errno));
      napster_quit ();
    }
    
    //    log_debug (DEBUG_NAPSTER, "c= %02x %c", c&255, c);

    if (c != '\n') {
      *(s+cnt) = c;
      cnt++;
    }
  }
}

static void *read_nap_process (void *dummy)
{
  char buf[256];

  log_debug (DEBUG_NAPSTER, "read_nap_process: starting");



  while (1) {
    read_nap_fifo (buf, sizeof(buf));

    switch (buf[1]) {
      
      /* Message */
    case 'M':
      /* Format is '@M y message' */
      lcd_write_center (buf[3]-'0', &buf[5], ATTR_NORMAL);
      break;
      
      /* Song */
    case 'S':
      strcpy (song_name, &buf[3]);
      pthread_mutex_lock (&scroll_mutex);
      lcd_init_hscroll ();
      scrolled_str = song_name;
      pthread_mutex_unlock (&scroll_mutex);

      log_debug (DEBUG_NAPSTER, "buf= '%s' ==> '%s'", buf, scrolled_str);
      break;

      /* Length in bytes */
    case 'L':
      lcd_write_center (2, &buf[3], ATTR_NORMAL);
      break;

      /* bit-rate + conn type */
    case 'R':
      lcd_write_center (3, &buf[3], ATTR_NORMAL);
      break;

    default:
      log_debug (DEBUG_NAPSTER, "read_nap_process: %s", buf);
      break;
    }
  }
}


/* Default input handler */
static void napster_default_handler ()
{
  omm_api_event ev = omm_api_get_current_event();

  log_debug (DEBUG_NAPSTER, "napster_default_handler: got '%c' in state %02x", ev, omm_state);
}

static void usage (char *s)
{
  fprintf (stderr, "Usage: %s\n", s);
  exit (1);
}

/* Main function */
main (int ac, char **av)
{
  char *cp, *new_omm_dir, buf[MAXPATHLEN];

  progname = (char*)basename(av[0]);
  debuglevel = DEFAULT_DEBUGLEVEL;

  if ((new_omm_dir = getenv ("OMMDIR")))
    omm_dir = new_omm_dir;

  init_debug ();

  /* Get options */
  while (--ac) {
    if ((cp = *++av) == NULL)
      break;
    if (*cp == '-' && *++cp) {
      switch(*cp) {

      default: 
	usage(progname);
	break;
      }
    }
    else
      break;
  }

  log_debug (DEBUG_NAPSTER, "starting.");


  if (pipe (pipe_out) < 0) {
    log_err ("pipe_out: %s", strerror(errno));
    exit (1);
  }

  if ((fd_napio = open_fifo (NAPIO)) < 0)
    exit (1);

  /* Init API */
  omm_api_init (napster_event_handlers, NAPSTER_FIFO, FALSE);

  lcd_clear_screen ();
  lcd_write (0, 0, "NAP", ATTR_NORMAL);

  /* Create str display process */
  pthread_mutex_init (&scroll_mutex, NULL);
  if (pthread_create (&scrolled_thread, &thread_attr, display_scrolled_process, NULL) != 0)
    log_err ("pthread_create error for display_scrolled_process");


  if (pthread_create (&read_nap_thread, &thread_attr, read_nap_process, NULL) != 0)
    log_err ("pthread_create error for read_nap_process");

  if (!(nap_pid = fork ())) {
    dup2 (pipe_out[0], 0);
    close (pipe_out[0]);

    putenv ("LINES=4");

    sprintf (buf, "%s/conf/nap.conf", omm_dir);

    if (execlp (NAP, NAP, "-f", buf, "-t", "-b", "-o", NAPIO, NULL) < 0) {
      log_err ("execlp: %s: %s", NAP, strerror(errno));
      exit (1);
    }
  }
  else {
    signal (SIGINT, got_interrupt);
    signal (SIGQUIT, got_interrupt);
    signal (SIGTERM, got_interrupt);


    /* Main loop */
    omm_api_main_loop ();

    /* End */
    log_debug (DEBUG_NAPSTER, "exiting.");
    close_debug ();
  }
}

